#!/usr/bin/env python

import os
import hashlib
import contextlib
import sqlite3

import logging
log = logging.getLogger('gglsbl')
log.addHandler(logging.NullHandler())

from gglsbl.utils import to_hex

class ThreatList(object):
    """Represents threat list name.
    """
    def __init__(self, threat_type, platform_type, threat_entry_type):
        self.threat_type = threat_type
        self.platform_type = platform_type
        self.threat_entry_type = threat_entry_type

    @classmethod
    def from_api_entry(cls, entry):
        return cls(entry['threatType'], entry['platformType'], entry['threatEntryType'])

    def as_tuple(self):
        return (self.threat_type, self.platform_type, self.threat_entry_type)

    def __repr__(self):
        return '/'.join(self.as_tuple())


class HashPrefixList(object):
    """Wrapper object for threat list data.
    """
    def __init__(self, prefix_size, raw_hashes):
        self.prefix_size = prefix_size
        self.raw_hashes = raw_hashes

    def __len__(self):
        return int(len(self.raw_hashes) / self.prefix_size)

    def __iter__(self):
        n = self.prefix_size
        return (self.raw_hashes[i:i+n] for i in range(0, len(self.raw_hashes), n))


class SqliteStorage(object):
    """Storage abstraction for local GSB cache"""
    def __init__(self, db_path):
        self.db_path = db_path
        do_init_db = not os.path.isfile(db_path)
        log.info('Opening SQLite DB {}'.format(db_path))
        self.db = sqlite3.connect(db_path)
        if do_init_db:
            log.info('SQLite DB does not exist, initializing')
            self.init_db()
        self.db.cursor().execute('PRAGMA synchronous = 0')

    @contextlib.contextmanager
    def get_cursor(self):
        dbc = self.db.cursor()
        try:
            yield dbc
        finally:
            dbc.close()

    def init_db(self):
        with self.get_cursor() as dbc:
            dbc.execute(
            """CREATE TABLE threat_list (
                threat_type character varying(128) NOT NULL,
                platform_type character varying(128) NOT NULL,
                threat_entry_type character varying(128) NOT NULL,
                client_state character varying(42),
                timestamp timestamp without time zone DEFAULT current_timestamp,
                PRIMARY KEY (threat_type, platform_type, threat_entry_type)
                )"""
            )
            dbc.execute(
            """CREATE TABLE full_hash (
                value BLOB NOT NULL,
                threat_type character varying(128) NOT NULL,
                platform_type character varying(128) NOT NULL,
                threat_entry_type character varying(128) NOT NULL,
                downloaded_at timestamp without time zone DEFAULT current_timestamp,
                expires_at timestamp without time zone NOT NULL DEFAULT current_timestamp,
                malware_threat_type varchar(32),
                PRIMARY KEY (value, threat_type, platform_type, threat_entry_type)
                )"""
            )
            dbc.execute(
            """CREATE TABLE hash_prefix (
                value BLOB NOT NULL,
                cue character varying(4) NOT NULL,
                threat_type character varying(128) NOT NULL,
                platform_type character varying(128) NOT NULL,
                threat_entry_type character varying(128) NOT NULL,
                timestamp timestamp without time zone DEFAULT current_timestamp,
                negative_expires_at timestamp without time zone NOT NULL DEFAULT current_timestamp,
                PRIMARY KEY (value, threat_type, platform_type, threat_entry_type),
                FOREIGN KEY(threat_type, platform_type, threat_entry_type)
                    REFERENCES threat_list(threat_type, platform_type, threat_entry_type)
                    ON DELETE CASCADE
                )"""
            )
            dbc.execute(
                """CREATE INDEX idx_hash_prefix_cue ON hash_prefix (cue)"""
            )
            dbc.execute(
                """CREATE INDEX idx_full_hash_expires_at ON full_hash (expires_at)"""
            )
            dbc.execute(
                """CREATE INDEX idx_full_hash_value ON full_hash (value)"""
            )
        self.db.commit()

    def lookup_full_hashes(self, hash_values):
        "Query DB to see if hash is blacklisted"
        q = '''SELECT threat_type,platform_type,threat_entry_type, expires_at < current_timestamp AS has_expired
                FROM full_hash WHERE value IN ({})
        '''
        output = []
        with self.get_cursor() as dbc:
            dbc.execute(q.format(','.join(['?']*len(hash_values))), [sqlite3.Binary(hv) for hv in hash_values])
            for h in dbc.fetchall():
                threat_type, platform_type, threat_entry_type, has_expired = h
                threat_list = ThreatList(threat_type, platform_type, threat_entry_type)
                output.append((threat_list, has_expired))
        return output

    def lookup_hash_prefix(self, cues):
        """Lookup hash prefixes by cue (first 4 bytes of hash)

        Returns a tuple of (threat_list, value, negative_cache_expired).
        """
        q = '''SELECT value,threat_type,platform_type,threat_entry_type,
                    negative_expires_at < current_timestamp AS negative_cache_expired
                FROM hash_prefix WHERE cue IN ({})
        '''
        output = []
        with self.get_cursor() as dbc:
            dbc.execute(q.format(','.join(['?'] * len(cues))), cues)
            for h in dbc.fetchall():
                value, threat_type, platform_type, threat_entry_type, negative_cache_expired = h
                threat_list = ThreatList(threat_type, platform_type, threat_entry_type)
                output.append((threat_list, bytes(value), negative_cache_expired))
        return output

    def store_full_hash(self, threat_list, hash_value, cache_duration, malware_threat_type):
        "Store full hash found for the given hash prefix"

        log.info('Storing full hash {} to list {} with cache duration {}'.format(to_hex(hash_value), str(threat_list), cache_duration))
        qi = '''INSERT OR IGNORE INTO full_hash
                    (value, threat_type, platform_type, threat_entry_type, malware_threat_type, downloaded_at)
                VALUES
                    (?, ?, ?, ?, ?, current_timestamp)
        '''
        qu = "UPDATE full_hash SET expires_at=datetime(current_timestamp, '+{} SECONDS') \
            WHERE value=? AND threat_type=? AND platform_type=? AND threat_entry_type=?"

        i_parameters = [sqlite3.Binary(hash_value), threat_list.threat_type,
                    threat_list.platform_type, threat_list.threat_entry_type, malware_threat_type]
        u_parameters = [sqlite3.Binary(hash_value), threat_list.threat_type,
                    threat_list.platform_type, threat_list.threat_entry_type]

        with self.get_cursor() as dbc:
            dbc.execute(qi, i_parameters)
            dbc.execute(qu.format(int(cache_duration)), u_parameters)
        self.db.commit()

    def delete_hash_prefix_list(self, threat_list):
        q = '''DELETE FROM hash_prefix
                    WHERE threat_type=? AND platform_type=? AND threat_entry_type=?
        '''
        parameters = [threat_list.threat_type, threat_list.platform_type, threat_list.threat_entry_type]
        with self.get_cursor() as dbc:
            dbc.execute(q, parameters)
        self.db.commit()

    def cleanup_full_hashes(self, keep_expired_for=60*60*12):
        """Remove long expired full_hash entries.
        """
        q = '''DELETE FROM full_hash WHERE expires_at=datetime(current_timestamp, '{} SECONDS')
        '''
        log.info('Cleaning up full_hash entries expired more than {} seconds ago.'.format(keep_expired_for))
        with self.get_cursor() as dbc:
            dbc.execute(q.format(int(keep_expired_for)))
        self.db.commit()

    def update_hash_prefix_expiration(self, threat_list, prefix_value, negative_cache_duration):
        q = "UPDATE hash_prefix SET negative_expires_at=datetime(current_timestamp, '+{} SECONDS') \
            WHERE value=? AND threat_type=? AND platform_type=? AND threat_entry_type=?"
        parameters = [sqlite3.Binary(prefix_value), threat_list.threat_type,
                    threat_list.platform_type, threat_list.threat_entry_type]
        with self.get_cursor() as dbc:
            dbc.execute(q.format(int(negative_cache_duration)), parameters)
        self.db.commit()

    def get_threat_lists(self):
        """Get a list of known threat lists including clientState values.
        """
        q = '''SELECT threat_type,platform_type,threat_entry_type,client_state FROM threat_list'''
        output = []
        with self.get_cursor() as dbc:
            dbc.execute(q)
            for h in dbc.fetchall():
                threat_type, platform_type, threat_entry_type, client_state = h
                threat_list = ThreatList(threat_type, platform_type, threat_entry_type)
                output.append((threat_list, client_state))
        return output

    def add_threat_list(self, threat_list):
        """Add threat list entry if it does not exist.
        """
        q = '''INSERT OR IGNORE INTO threat_list
                    (threat_type, platform_type, threat_entry_type, timestamp)
                VALUES
                    (?, ?, ?, current_timestamp)
        '''
        params = [threat_list.threat_type, threat_list.platform_type, threat_list.threat_entry_type]
        with self.get_cursor() as dbc:
            dbc.execute(q, params)
        self.db.commit()

    def delete_threat_list(self, threat_list):
        """Delete threat list entry.
        """
        log.info('Deleting cached threat list "{}"'.format(repr(threat_list)))
        q = '''DELETE FROM threat_list
                    WHERE threat_type=? AND platform_type=? AND threat_entry_type=?
        '''
        params = [threat_list.threat_type, threat_list.platform_type, threat_list.threat_entry_type]
        with self.get_cursor() as dbc:
            dbc.execute(q, params)
        self.db.commit()

    def update_threat_list_client_state(self, threat_list, client_state):
        log.info('Setting client_state of threat list {} to {}'.format(str(threat_list), client_state))
        q = '''UPDATE threat_list SET timestamp=current_timestamp, client_state=?
            WHERE threat_type=? AND platform_type=? AND threat_entry_type=?'''
        params = [client_state, threat_list.threat_type, threat_list.platform_type, threat_list.threat_entry_type]
        with self.get_cursor() as dbc:
            dbc.execute(q, params)
        self.db.commit()

    def hash_prefix_list_checksum(self, threat_list):
        """Returns SHA256 checksum for alphabetically-sorted concatenated list of hash prefixes
        """
        q = '''SELECT value FROM hash_prefix
                WHERE threat_type=? AND platform_type=? AND threat_entry_type=?
                ORDER BY value
        '''
        params = [threat_list.threat_type, threat_list.platform_type, threat_list.threat_entry_type]
        with self.get_cursor() as dbc:
            dbc.execute(q, params)
            all_hashes = b''.join([ bytes(h[0]) for h in dbc.fetchall() ])
            checksum = hashlib.sha256(all_hashes).digest()
        return checksum

    def populate_hash_prefix_list(self, threat_list, hash_prefix_list):
        log.info('Storing {} entries of hash prefix list {}'.format(len(hash_prefix_list), str(threat_list)))
        q = '''INSERT INTO hash_prefix
                    (value, cue, threat_type, platform_type, threat_entry_type, timestamp)
                VALUES
                    (?, ?, ?, ?, ?, current_timestamp)
        '''
        with self.get_cursor() as dbc:
            records = [[sqlite3.Binary(prefix_value), to_hex(prefix_value[0:4]), threat_list.threat_type,
                        threat_list.platform_type, threat_list.threat_entry_type] for prefix_value in hash_prefix_list]
            dbc.executemany(q, records)
        #self.db.commit()

    def get_hash_prefix_values_to_remove(self, threat_list, indices):
        log.info('Removing {} records from threat list "{}"'.format(len(indices), str(threat_list)))
        indices = set(indices)
        q = '''SELECT value FROM hash_prefix
                WHERE threat_type=? AND platform_type=? AND threat_entry_type=?
                ORDER BY value
        '''
        params = [threat_list.threat_type, threat_list.platform_type, threat_list.threat_entry_type]
        values_to_remove = []
        with self.get_cursor() as dbc:
            dbc.execute(q, params)
            i = 0
            for h in dbc.fetchall():
                v = bytes(h[0])
                if i in indices:
                    values_to_remove.append(v)
                i += 1
        return values_to_remove

    def remove_hash_prefix_indices(self, threat_list, indices):
        """Remove records matching idices from a lexicographically-sorted local threat list.
        """
        batch_size = 40
        q = '''DELETE FROM hash_prefix
                WHERE threat_type=? AND platform_type=? AND threat_entry_type=? AND value IN ({})
        '''
        prefixes_to_remove = self.get_hash_prefix_values_to_remove(threat_list, indices)
        with self.get_cursor() as dbc:
            for i in range(0, len(prefixes_to_remove), batch_size):
                remove_batch = prefixes_to_remove[i:(i+batch_size)]
                params = [threat_list.threat_type, threat_list.platform_type, threat_list.threat_entry_type] + \
                                [sqlite3.Binary(b) for b in remove_batch]
                dbc.execute(q.format(','.join(['?'] * len(remove_batch))), params)

    def total_cleanup(self):
        "Reset local cache"
        with self.get_cursor() as dbc:
            q = 'DROP TABLE threat_list'
            dbc.execute(q)
            q = 'DROP TABLE hash_prefix'
            dbc.execute(q)
            q = 'DROP TABLE full_prefix'
            dbc.execute(q)
        self.db.commit()
        self.init_db()



#!/usr/bin/env python

from functools import wraps

try:
    import urllib, urlparse
except ImportError:
    import urllib.parse as urllib
    from urllib import parse as urlparse

import struct
import time
import posixpath
import re
import hashlib
import socket
import random
from base64 import b64encode, b64decode

try:
    from googleapiclient.discovery import build
    from googleapiclient.errors import HttpError
except ImportError:
    from apiclient.discovery import build
    from apiclient.errors import HttpError

import logging
log = logging.getLogger('gglsbl')
log.addHandler(logging.NullHandler())

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


_fail_count = 0
def autoretry(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        global _fail_count
        while True:
            try:
                r = func(*args, **kwargs)
                _fail_count = 0
                return r
            except HttpError as e:
                if not (hasattr(e, 'resp') and 'status' in e.resp \
                        and e.resp['status'].isdigit and int(e.resp['status']) >= 500):
                    raise # we do not want to retry auth errors etc.
                _fail_count += 1
                wait_for = min(2**(_fail_count - 1) * 15 * 60 * (1 + random.random()), 24*60*60)
                log.exception('Call Failed for {} time(s). Retrying in {} seconds: {}'.format(_fail_count, wait_for, str(e)))
                time.sleep(wait_for)
            except socket.error as e:
                transient_error_wait = 2
                log.exception('Socket error, retrying in {} seconds.'.format(transient_error_wait))
                time.sleep(transient_error_wait)
    return wrapper


class SafeBrowsingApiClient(object):
    def __init__(self, developer_key, client_id='python-gglsbl', client_version=__version__, discard_fair_use_policy=True):
        self.client_id = client_id
        self.client_version = client_version
        self.discard_fair_use_policy = discard_fair_use_policy
        if self.discard_fair_use_policy:
            log.warn('Circumventing request frequency throttling is against Safe Browsing API policy.')
        self.service = build('safebrowsing', 'v4', developerKey=developer_key)
        self.next_request_no_sooner_than = None

    def set_wait_duration(self, minimum_wait_duration):
        if self.discard_fair_use_policy:
            return
        if minimum_wait_duration is None:
            self.next_request_no_sooner_than = None
            return
        self.next_request_no_sooner_than = time.time() + float(minimum_wait_duration.rstrip('s'))

    def fair_use_delay(self):
        if self.next_request_no_sooner_than is not None:
            sleep_for = self.next_request_no_sooner_than - time.time()
            log.info('Sleeping for {} seconds until next request.'.format(sleep_for))
            time.sleep(sleep_for)

    @autoretry
    def get_threats_lists(self):
        """Retrieve all available threat lists
        """
        response = self.service.threatLists().list().execute()
        self.set_wait_duration(response.get('minimumWaitDuration'))
        return response['threatLists']

    @autoretry
    def get_threats_update(self, client_state):
        """Fetch hash prefixes update for given threat list.

        client_state is a dict which looks like {(threatType, platformType, threatEntryType): clientState}
        """
        request_body = {
                "client": {
                "clientId":       self.client_id,
                "clientVersion":  self.client_version,
            },
            "listUpdateRequests": []
        }
        for (threat_type, platform_type, threat_entry_type), current_state in client_state.items():
            request_body['listUpdateRequests'].append(
                {
                    "threatType":      threat_type,
                    "platformType":    platform_type,
                    "threatEntryType": threat_entry_type,
                    "state":           current_state,
                    "constraints": {
                        "supportedCompressions": ["RAW"]
                    }
                }
            )
        response = self.service.threatListUpdates().fetch(body=request_body).execute()
        self.set_wait_duration(response.get('minimumWaitDuration'))
        return response['listUpdateResponses']

    @autoretry
    def get_full_hashes(self, prefixes, client_state):
        """Find full hashes matching hash prefixes.

        client_state is a dict which looks like {(threatType, platformType, threatEntryType): clientState}
        """
        request_body = {
          "client": {
            "clientId":      self.client_id,
            "clientVersion": self.client_version,
          },
          "clientStates": [],
          "threatInfo": {
            "threatTypes":      [],
            "platformTypes":    [],
            "threatEntryTypes": [],
            "threatEntries": [],
          }
        }
        for prefix in prefixes:
            request_body['threatInfo']['threatEntries'].append({"hash": b64encode(prefix).decode()})
        for ((threatType, platformType, threatEntryType), clientState) in client_state.items():
            request_body['clientStates'].append(clientState)
            if threatType not in request_body['threatInfo']['threatTypes']:
                request_body['threatInfo']['threatTypes'].append(threatType)
            if platformType not in request_body['threatInfo']['platformTypes']:
                request_body['threatInfo']['platformTypes'].append(platformType)
            if threatEntryType not in request_body['threatInfo']['threatEntryTypes']:
                request_body['threatInfo']['threatEntryTypes'].append(threatEntryType)
        response = self.service.fullHashes().find(body=request_body).execute()
        self.set_wait_duration(response.get('minimumWaitDuration'))
        return response

class URL(object):
    "URL representation suitable for lookup"
    def __init__(self, url):
        self.url = str(url)

    @property
    def hashes(self):
        "Hashes of all possible permutations of the URL in canonical form"
        for url_variant in self.url_permutations(self.canonical):
            url_hash = self.digest(url_variant)
            yield url_hash

    @property
    def canonical(self):
        "Convert URL to its canonical form"
        def full_unescape(u):
            uu = urllib.unquote(u)
            if uu == u:
                return uu
            else:
                return full_unescape(uu)
        def quote(s):
            safe_chars = '!"$&\'()*+,-./:;<=>?@[\\]^_`{|}~'
            return urllib.quote(s, safe=safe_chars)
        url = self.url.strip()
        url = url.replace('\n', '').replace('\r', '').replace('\t', '')
        url = url.split('#', 1)[0]
        url = quote(full_unescape(url))
        url_parts = urlparse.urlsplit(url)
        if not url_parts[0]:
            url = 'http://%s' % url
            url_parts = urlparse.urlsplit(url)
        protocol = url_parts.scheme
        host = full_unescape(url_parts.hostname)
        path = full_unescape(url_parts.path)
        query = url_parts.query
        if not query and '?' not in url:
            query = None
        if not path:
            path = '/'
        has_trailing_slash = (path[-1] == '/')
        path = posixpath.normpath(path).replace('//', '/')
        if has_trailing_slash and path[-1] != '/':
            path = path + '/'
        user = url_parts.username
        port = url_parts.port
        host = host.strip('.')
        host = re.sub(r'\.+', '.', host).lower()
        if host.isdigit():
            try:
                host = socket.inet_ntoa(struct.pack("!I", int(host)))
            except:
                pass
        if host.startswith('0x') and '.' not in host:
            try:
                host = socket.inet_ntoa(struct.pack("!I", int(host, 16)))
            except:
                pass
        quoted_path = quote(path)
        quoted_host = quote(host)
        if port is not None:
            quoted_host = '%s:%s' % (quoted_host, port)
        canonical_url = '%s://%s%s' % (protocol, quoted_host, quoted_path)
        if query is not None:
            canonical_url = '%s?%s' % (canonical_url, query)
        return canonical_url

    @staticmethod
    def url_permutations(url):
        """Try all permutations of hostname and path which can be applied
        to blacklisted URLs"""
        def url_host_permutations(host):
            if re.match(r'\d+\.\d+\.\d+\.\d+', host):
                yield host
                return
            parts = host.split('.')
            l = min(len(parts),5)
            if l > 4:
                yield host
            for i in range(l-1):
                yield '.'.join(parts[i-l:])
        def url_path_permutations(path):
            yield path
            query = None
            if '?' in path:
                path, query =  path.split('?', 1)
            if query is not None:
                yield path
            path_parts = path.split('/')[0:-1]
            curr_path = ''
            for i in range(min(4, len(path_parts) )):
                curr_path = curr_path + path_parts[i] + '/'
                yield curr_path
        protocol, address_str = urllib.splittype(url)
        host, path = urllib.splithost(address_str)
        user, host = urllib.splituser(host)
        host, port = urllib.splitport(host)
        host = host.strip('/')
        seen_permutations = set()
        for h in url_host_permutations(host):
            for p in url_path_permutations(path):
                u = '%s%s' % (h, p)
                if u not in seen_permutations:
                    yield u
                    seen_permutations.add(u)

    @staticmethod
    def digest(url):
        "Hash the URL"
        return hashlib.sha256(url.encode('utf-8')).digest()


if __name__ == '__main__':
    from pprint import pprint
    c = SafeBrowsingApiClient('AIzaSyATpqLltciaMve61Wywb5yNDA8D8BvXEn4')
    r = c.get_threats_lists()
    pprint(r)



import unittest

from gglsbl.protocol import URL
from gglsbl import SafeBrowsingList

class SafeBrowsingListTestCase(unittest.TestCase):
    def setUp(self):
        self.canonical_urls = {
            "http://host/%25%32%35": "http://host/%25",
            "http://host/%25%32%35%25%32%35": "http://host/%25%25",
            "http://host/%2525252525252525": "http://host/%25",
            "http://host/asdf%25%32%35asd": "http://host/asdf%25asd",
            "http://host/%%%25%32%35asd%%": "http://host/%25%25%25asd%25%25",
            "http://www.google.com/": "http://www.google.com/",
            "http://%31%36%38%2e%31%38%38%2e%39%39%2e%32%36/%2E%73%65%63%75%72%65/%77%77%77%2E%65%62%61%79%2E%63%6F%6D/": "http://168.188.99.26/.secure/www.ebay.com/",
            "http://195.127.0.11/uploads/%20%20%20%20/.verify/.eBaysecure=updateuserdataxplimnbqmn-xplmvalidateinfoswqpcmlx=hgplmcx/": "http://195.127.0.11/uploads/%20%20%20%20/.verify/.eBaysecure=updateuserdataxplimnbqmn-xplmvalidateinfoswqpcmlx=hgplmcx/",
            "http://host%23.com/%257Ea%2521b%2540c%2523d%2524e%25f%255E00%252611%252A22%252833%252944_55%252B": "http://host%23.com/~a!b@c%23d$e%25f^00&11*22(33)44_55+",
            "http://3279880203/blah": "http://195.127.0.11/blah",
            "http://0xc37f000b/blah": "http://195.127.0.11/blah",
            "http://www.google.com/blah/..": "http://www.google.com/",
            "www.google.com/": "http://www.google.com/",
            "www.google.com": "http://www.google.com/",
            "http://www.evil.com/blah#frag": "http://www.evil.com/blah",
            "http://www.GOOgle.com/": "http://www.google.com/",
            "google.com": "http://google.com/",
            "http://www.google.com.../": "http://www.google.com/",
            "http://www.google.com/foo\tbar\rbaz\n2":"http://www.google.com/foobarbaz2",
            "http://www.google.com/q?": "http://www.google.com/q?",
            "http://www.google.com/q?r?": "http://www.google.com/q?r?",
            "http://www.google.com/q?r?s": "http://www.google.com/q?r?s",
            "http://evil.com/foo#bar#baz": "http://evil.com/foo",
            "http://evil.com/foo;": "http://evil.com/foo;",
            "http://evil.com/foo?bar;": "http://evil.com/foo?bar;",
            "http://\x01\x80.com/": "http://%01%80.com/",
            "http://notrailingslash.com": "http://notrailingslash.com/",
            "http://www.gotaport.com:1234/": "http://www.gotaport.com:1234/",
            "  http://www.google.com/  ": "http://www.google.com/",
            "http:// leadingspace.com/": "http://%20leadingspace.com/",
            "http://%20leadingspace.com/": "http://%20leadingspace.com/",
            "%20leadingspace.com/": "http://%20leadingspace.com/",
            "https://www.securesite.com/": "https://www.securesite.com/",
            "http://host.com/ab%23cd": "http://host.com/ab%23cd",
            "http://host.com//twoslashes?more//slashes": "http://host.com/twoslashes?more//slashes",
            "http://www.wtp101.com/bk?redir=http%3A%2F%2Ftags.bluekai.com%2Fsite%2F2750%3Fid%3D%3CPARTNER_UUID%3E%0D%0A%26redir%3Dhttp%3A%2F%2Fwww.wtp101.com%2Fpush%2Fbluekai%3Fxid%3D%24BK_UUID": "http://www.wtp101.com/bk?redir=http://tags.bluekai.com/site/2750?id=<PARTNER_UUID>%0D%0A&redir=http://www.wtp101.com/push/bluekai?xid=$BK_UUID",
        }

        self.url_permutations = {
            'http://a.b.c/1/2.html?param=1': [
                'a.b.c/1/2.html?param=1',
                'a.b.c/1/2.html',
                'a.b.c/',
                'a.b.c/1/',
                'b.c/1/2.html?param=1',
                'b.c/1/2.html',
                'b.c/',
                'b.c/1/',
            ],
            'http://a.b.c/1/2/?param=1': [
                'a.b.c/1/2/?param=1',
                'a.b.c/1/2/',
                'a.b.c/',
                'a.b.c/1/',
                'b.c/1/2/?param=1',
                'b.c/1/2/',
                'b.c/',
                'b.c/1/',
            ],
            'http://1.2.3.4/1/2.html?param=1': [
                '1.2.3.4/1/2.html?param=1',
                '1.2.3.4/1/2.html',
                '1.2.3.4/',
                '1.2.3.4/1/',
            ],
            'http://a.b.c/1/2/3/4/5/6/7.html?param=1': [
                'a.b.c/1/2/3/4/5/6/7.html?param=1',
                'a.b.c/1/2/3/4/5/6/7.html',
                'a.b.c/',
                'a.b.c/1/',
                'a.b.c/1/2/',
                'a.b.c/1/2/3/',
                'b.c/1/2/3/4/5/6/7.html?param=1',
                'b.c/1/2/3/4/5/6/7.html',
                'b.c/',
                'b.c/1/',
                'b.c/1/2/',
                'b.c/1/2/3/',
            ],
            'ttp://a.b.c.d.e.f.g/1.html': [
                'a.b.c.d.e.f.g/1.html',
                'a.b.c.d.e.f.g/',
                'c.d.e.f.g/1.html',
                'c.d.e.f.g/',
                'd.e.f.g/1.html',
                'd.e.f.g/',
                'e.f.g/1.html',
                'e.f.g/',
                'f.g/1.html',
                'f.g/',
            ],
            'http://a.b/': [
                'a.b/',
            ],
        }

    def test_canonicalize(self):
        for nu, cu in self.canonical_urls.items():
            self.assertEqual(URL(nu).canonical, cu)

    def test_permutations(self):
        for k,v in self.url_permutations.items():
            p = list(URL.url_permutations(k))
            self.assertEqual(p, v)




#!/usr/bin/env python

__all__ = [
    'SafeBrowsingList'
]

from gglsbl.client import SafeBrowsingList

from gglsbl._version import get_versions
__version__ = get_versions()['version']
del get_versions



#!/usr/bin/env python

"""Keeps local Google Safe Browsing cache in sync.

Accessing Google Safe Browsing API requires API key, you can find
more info on getting it here:
https://developers.google.com/safe-browsing/lookup_guide#GettingStarted

"""

import argparse
import json
import sys
import time

from gglsbl import SafeBrowsingList

import logging
log = logging.getLogger('gglsbl')
log.setLevel(logging.DEBUG)


def setupArgsParser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--api-key',
                default=None,
                required=True,
                help='Safe Browsing API key [REQUIRED]')
    parser.add_argument('--db-path',
                default='/tmp/gsb_v4.db',
                help='Path to SQLite DB')
    parser.add_argument('--log',
                default=None,
                help='Path to log file, by default log to STDERR')
    parser.add_argument('--check-url',
                default=None,
                help='Check if URL is in black list and exit')
    parser.add_argument('--debug',
                default=False,
                action = 'store_true',
                help='Show debug output')
    parser.add_argument('--onetime',
                default=False,
                action = 'store_true',
                help='Run blacklists sync only once with reduced delays')
    return parser

def setupLogger(log_file, debug):
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    lh = log_file is None and logging.StreamHandler() or logging.FileHandler(log_file)
    lh.setLevel(debug and logging.DEBUG or logging.INFO)
    lh.setFormatter(formatter)
    log = logging.getLogger('gglsbl')
    log.addHandler(lh)

def run_sync(sbl):
    try:
        sbl.update_hash_prefix_cache()
    except (KeyboardInterrupt, SystemExit):
        log.info('Shutting down')
        sys.exit(0)
    except:
        log.exception('Failed to synchronize with GSB service')
        time.sleep(3)

def main():
    args_parser = setupArgsParser()
    args = args_parser.parse_args()
    setupLogger(args.log, args.debug)
    if args.check_url:
        sbl = SafeBrowsingList(args.api_key, db_path=args.db_path)
        bl = sbl.lookup_url(args.check_url)
        if bl is None:
            print('{} is not blacklisted'.format(args.check_url))
        else:
            print('{} is blacklisted in {}'.format(args.check_url, bl))
        sys.exit(0)
    if args.onetime:
        sbl = SafeBrowsingList(args.api_key, db_path=args.db_path, discard_fair_use_policy=True)
        run_sync(sbl)
    else:
        sbl = SafeBrowsingList(args.api_key, db_path=args.db_path)
        while True:
            run_sync(sbl)

if __name__ == '__main__':
    main()





# This file was generated by 'versioneer.py' (0.17) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2017-05-14T11:54:35+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "4f0227ca58868e81cbc16efb9940ed0ea5f9a5d1",
 "version": "v1.4.4"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



#!/usr/bin/env python

from base64 import b64encode, b64decode

import logging
log = logging.getLogger('gglsbl')
log.addHandler(logging.NullHandler())

from gglsbl.utils import to_hex
from gglsbl.protocol import SafeBrowsingApiClient, URL
from gglsbl.storage import SqliteStorage, ThreatList, HashPrefixList


class SafeBrowsingList(object):
    """Interface for Google Safe Browsing API

    supporting partial update of the local cache.
    https://developers.google.com/safe-browsing/v4/
    """

    def __init__(self, api_key, db_path='/tmp/gsb_v4.db', discard_fair_use_policy=False,
                        platforms = None):
        """Constructor.

        Args:
            api_key: string, a key for API authentication.
            db_path: string, path to SQLite DB file to store cached data.
            discard_fair_use_policy: boolean, disable request frequency throttling (only for testing).
            platforms: list, threat lists to look up, default includes all platforms.
        """
        self.api_client = SafeBrowsingApiClient(api_key, discard_fair_use_policy=discard_fair_use_policy)
        self.storage = SqliteStorage(db_path)
        self.platforms = platforms

    def _verify_threat_list_checksum(self, threat_list, remote_checksum):
        local_checksum = self.storage.hash_prefix_list_checksum(threat_list)
        return remote_checksum == local_checksum

    def update_hash_prefix_cache(self):
        """Update locally cached threat lists.
        """
        self.api_client.fair_use_delay()
        self.storage.cleanup_full_hashes()
        threat_lists_to_remove = dict()
        for ts, cs in self.storage.get_threat_lists():
            threat_lists_to_remove[repr(ts)] = ts
        threat_lists = self.api_client.get_threats_lists()
        for entry in threat_lists:
            threat_list = ThreatList.from_api_entry(entry)
            if self.platforms is None or threat_list.platform_type in self.platforms:
                self.storage.add_threat_list(threat_list)
                try:
                    del threat_lists_to_remove[repr(threat_list)]
                except KeyError:
                    pass
        for ts in threat_lists_to_remove.values():
            self.storage.delete_hash_prefix_list(ts)
            self.storage.delete_threat_list(ts)
        del threat_lists_to_remove

        self.api_client.fair_use_delay()
        threat_lists = self.storage.get_threat_lists()
        client_state = dict([(t.as_tuple(), s) for t,s in threat_lists])
        for response in self.api_client.get_threats_update(client_state):
            response_threat_list = ThreatList(response['threatType'], response['platformType'], response['threatEntryType'])
            if response['responseType'] == 'FULL_UPDATE':
                self.storage.delete_hash_prefix_list(response_threat_list)
            for r in response.get('removals', []):
                self.storage.remove_hash_prefix_indices(response_threat_list, r['rawIndices']['indices'])
            for a in response.get('additions', []):
                hash_prefix_list = HashPrefixList(a['rawHashes']['prefixSize'], b64decode(a['rawHashes']['rawHashes']))
                self.storage.populate_hash_prefix_list(response_threat_list, hash_prefix_list)
            expected_checksum = b64decode(response['checksum']['sha256'])
            if self._verify_threat_list_checksum(response_threat_list, expected_checksum):
                log.info('Local cache checksum matches the server: {}'.format(to_hex(expected_checksum)))
                self.storage.update_threat_list_client_state(response_threat_list, response['newClientState'])
            else:
                raise Exception('Local cache checksum does not match the server: "{}". Consider removing {}'.format(to_hex(expected_checksum), self.storage.db_path))

    def _sync_full_hashes(self, hash_prefixes):
        """Download full hashes matching hash_prefixes.

        Also update cache expiration timetsamps.
        """
        threat_lists = self.storage.get_threat_lists()
        client_state = dict([(t.as_tuple(), s) for t,s in threat_lists])
        self.api_client.fair_use_delay()
        fh_response = self.api_client.get_full_hashes(hash_prefixes, client_state)

        # update negative cache for each hash prefix
        # store full hash (insert or update) with positive cache bumped up
        for m in fh_response['matches']:
            threat_list = ThreatList(m['threatType'], m['platformType'], m['threatEntryType'])
            hash_value = b64decode(m['threat']['hash'])
            cache_duration = int(m['cacheDuration'].rstrip('s'))
            malware_threat_type = None
            for metadata in m['threatEntryMetadata'].get('entries', []):
                k = b64decode(metadata['key'])
                v = b64decode(metadata['value'])
                if k == 'malware_threat_type':
                    malware_threat_type = v
            self.storage.store_full_hash(threat_list, hash_value, cache_duration, malware_threat_type)

        negative_cache_duration = int(fh_response['negativeCacheDuration'].rstrip('s'))
        for prefix_value in hash_prefixes:
            for threat_list in threat_lists:
                self.storage.update_hash_prefix_expiration(threat_list[0], prefix_value, negative_cache_duration)

    def lookup_url(self, url):
        """Look up specified URL in Safe Browsing threat lists."""
        if type(url) is not str:
            url = url.encode('utf8')
        url_hashes = URL(url).hashes
        list_names = self._lookup_hashes(url_hashes)
        if list_names:
            return list_names
        return None

    def _lookup_hashes(self, full_hashes):
        """Lookup URL hash in blacklists

        Returns names of lists it was found in.
        """
        full_hashes = list(full_hashes)
        cues = [to_hex(fh[0:4]) for fh in full_hashes]
        result = []
        try:
            matching_prefixes = {}
            matching_full_hashes = set()
            is_potential_threat = False
            # First lookup hash prefixes which match full URL hash
            for (threat_list, hash_prefix, negative_cache_expired) in self.storage.lookup_hash_prefix(cues):
                for full_hash in full_hashes:
                    if full_hash.startswith(hash_prefix):
                        is_potential_threat = True
                        # consider hash prefix negative cache as expired if it is expired in at least one threat list
                        matching_prefixes[hash_prefix] = matching_prefixes.get(hash_prefix, False) or negative_cache_expired
                        matching_full_hashes.add(full_hash)
            # if none matches, URL hash is clear
            if not is_potential_threat:
                return []
            # if there is non-expired full hash, URL is blacklisted
            matching_expired_threat_lists = set()
            for threat_list, has_expired in self.storage.lookup_full_hashes(matching_full_hashes):
                if has_expired:
                    matching_expired_threat_lists.add(threat_list)
                else:
                    result.append(threat_list)
            if result:
                return result

            # If there are no matching expired full hash entries
            # and negative cache is still current for all prefixes, consider it safe
            if len(matching_expired_threat_lists) == 0 and sum(map(int, matching_prefixes.values())) == 0:
                log.info('Negative cache hit.')
                return []

            # Now we can assume that there are expired matching full hash entries and/or
            # cache prefix entries with expired negative cache. Both require full hash sync.
            self._sync_full_hashes(matching_prefixes.keys())
            # Now repeat full hash lookup
            for threat_list, has_expired in self.storage.lookup_full_hashes(matching_full_hashes):
                if not has_expired:
                    result.append(threat_list)
        except:
            self.storage.db.rollback()
            raise
        return result



#!/usr/bin/env python

from setuptools import setup
import versioneer

import sys, os

setup(name='gglsbl',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Client library for Google Safe Browsing Update API v4",
    classifiers=[
        "Operating System :: POSIX",
        "Environment :: Console",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Internet",
        "Topic :: Security",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords='google safe browsing api client',
    author='Aleh Filipovich',
    author_email='aleh@vaolix.com',
    url='https://github.com/afilipovich/gglsbl',
    license='Apache2',
    packages=['gglsbl'],
    install_requires=['google-api-python-client'],
    scripts=['bin/gglsbl_client.py'],
)
